// Collections.h: defines a few general purpose collection classes
//

#pragma once

#undef DECL
#if defined(NOT_XPERT)
#define DECL
#else
#if defined(_AFXDLL) || defined(XPERT_EXE)
#ifdef UTILS_DLL
#define DECL _declspec(dllexport)
#else
#define DECL _declspec(dllimport)
#endif
#else
#define DECL 
#endif
#endif

///////////////////////////////////////////////////////////////////////////
// TMapStringToPtr class
// 
// Improves upon CMapStringToPtr by ignoring case when searching
// but retaining case when retrieving
//
class DECL TMapStringToPtr : public CObject
{
protected:
	// Association
	struct CAssoc
	{
		CAssoc* pNext;
		UINT nHashValue;  // needed for efficient iteration
		CString key;
		void* value;
	};

public:
// Construction
	/* explicit */ TMapStringToPtr(INT_PTR nBlockSize = 10);

// Attributes
	// number of elements
	INT_PTR GetCount() const { return m_nCount; };
	INT_PTR GetSize() const  { return m_nCount; };
	BOOL IsEmpty() const     { return m_nCount == 0; };

	// Lookup
	BOOL Lookup(LPCTSTR key, void*& rValue) const;
	BOOL LookupKey(LPCTSTR key, LPCTSTR& rKey) const;

// Operations
	// Lookup and add if not there
	void*& operator[](LPCTSTR key);

	// add a new (key, value) pair
	void SetAt(LPCTSTR key, void* newValue) { (*this)[key] = newValue; };

	// removing existing (key, ?) pair
	BOOL RemoveKey(LPCTSTR key);
	void RemoveAll();

	// iterating all (key, value) pairs
	POSITION GetStartPosition() const  { return (m_nCount == 0) ? NULL : BEFORE_START_POSITION; };
	void GetNextAssoc(POSITION& rNextPosition, CString& rKey, void*& rValue) const;

	// advanced features for derived classes
	UINT GetHashTableSize() const  { return m_nHashTableSize; };
	void InitHashTable(UINT hashSize, BOOL bAllocNow = TRUE);

// Overridables: special non-virtual (see map implementation for details)
	// Routine used to user-provided hash keys
	UINT HashKey(LPCTSTR key) const;

// Implementation
protected:
	CAssoc** m_pHashTable;
	UINT m_nHashTableSize;
	INT_PTR m_nCount;
	CAssoc* m_pFreeList;
	struct CPlex* m_pBlocks;
	INT_PTR m_nBlockSize;

	CAssoc* NewAssoc();
	void FreeAssoc(CAssoc*);
	CAssoc* GetAssocAt(LPCTSTR, UINT&, UINT&) const;

public:
	~TMapStringToPtr();
#ifdef _DEBUG
	void AssertValid() const;
#endif

protected:
	// local typedefs for CTypedPtrMap class template
	typedef CString BASE_KEY;
	typedef LPCTSTR BASE_ARG_KEY;
	typedef void* BASE_VALUE;
	typedef void* BASE_ARG_VALUE;
};

///////////////////////////////////////////////////////////////////////////
// TMapStringToRef class
// 
// Improves upon CMapStringToPtr by working with references instead of 
// ptrs and ignoring case while retaining case Type-safe by using templates.
//
template<class TYPE>
class TMapStringToRef : public TMapStringToPtr
{
public:
   BOOL Lookup(LPCTSTR key, TYPE*& V) const                    { return TMapStringToPtr::Lookup(key, (void *&) V);    };
   TYPE& operator[](LPCTSTR key)                               { return * (TYPE*&) TMapStringToPtr::operator[](key);  };
   void SetAt(LPCTSTR key, TYPE& V)                            { TMapStringToPtr::SetAt(key, &V);              };
   void GetNextAssoc(POSITION& P, CString& K, TYPE*& V) const  { TMapStringToPtr::GetNextAssoc(P, K, (void *&) V);    };
};


#define TMapStringToRefUpperCase TMapStringToRef

#if 0
///////////////////////////////////////////////////////////////////////////
// TMapStringToRefUpperCase class
// 
// Version of TMapStringToRef that uppercases all strings. 
// 
template<class TYPE>
class TMapStringToRefUpperCase : public CMapStringToPtr
{
public:
   BOOL Lookup(CString K, TYPE*& V) const                      { K.MakeUpper(); return CMapStringToPtr::Lookup(K, (void *&) V);    };
   TYPE& operator[](CString K)                                 { K.MakeUpper(); return * (TYPE*&) CMapStringToPtr::operator[](K);  };
   void SetAt(CString K, TYPE& V)                              { K.MakeUpper(); CMapStringToPtr::SetAt(K, &V);                     };
   void GetNextAssoc(POSITION& P, CString& K, TYPE*& V) const  { CMapStringToPtr::GetNextAssoc(P, K, (void *&) V);                 };
   BOOL RemoveKey(CString K)                                   { K.MakeUpper(); return CMapStringToPtr::RemoveKey(K);              };
};
#endif

///////////////////////////////////////////////////////////////////////////
// TObArray class
// 
// Improves upon CObArray by working with references instead of pointers. 
// Type-safe by using templates. 
// 
template<class TYPE>
class TObArray : public CObArray
{
public:
   TYPE& GetAt(int nIndex) const               { return (TYPE&) *CObArray::GetAt(nIndex);     };
   TYPE& ElementAt(int nIndex)                 { return (TYPE&) CObArray::ElementAt(nIndex);  };
   void SetAt(int nIndex, TYPE& newElement)    { CObArray::SetAt(nIndex, &newElement);        };
   void SetAtGrow(int nIndex, TYPE& newElement){ CObArray::SetAtGrow(nIndex, &newElement);    }
   int Add(TYPE& newElement)                   { return CObArray::Add(&newElement);           };
   int Append(const TObArray<TYPE>& src)       { return CObArray::Append(src);                };
   void Copy(const TObArray<TYPE>& src)        { CObArray::Copy(src);                         };
   void InsertAt(int nIndex, TYPE& newElement, int nCount = 1) { CObArray::InsertAt(nIndex, &newElement, nCount); };
   void InsertAt(int nStartIndex, TObArray<TYPE>& NewArray)    { CObArray::InsertAt(nStartIndex, &NewArray);      };
   TYPE& operator[](int nIndex) const          { return (TYPE&) *CObArray::operator[](nIndex);        };
   int GetSize() {return CObArray::GetSize();}
};

///////////////////////////////////////////////////////////////////////////
// TObList class
// 
// Improves upon CObList by working with references instead of pointers. 
// Type-safe by using templates. 
// 
template<class TYPE>
class TObList : public CObList
{
public:
   TYPE& GetHead()                                 {return (TYPE&)*CObList::GetHead();    }
   TYPE& GetTail()                                 {return (TYPE&)*CObList::GetTail();    }
   POSITION AddHead(TYPE& item)                    {return CObList::AddHead(&item);       }
   void AddHead(TObList<TYPE>& items)              {CObList::AddHead(items);              }
   POSITION AddTail(TYPE& item)                    {return CObList::AddTail(&item);       }
   void AddTail(TObList<TYPE>& items)              {CObList::AddTail(items);              }
   TYPE& RemoveHead()                              {return (TYPE&)*CObList::RemoveHead(); }
   TYPE& RemoveTail()                              {return (TYPE&)*CObList::RemoveTail(); }
   TYPE& GetNext(POSITION& pos)                    {return (TYPE&)*CObList::GetNext(pos); }
   TYPE& GetPrev(POSITION& pos)                    {return (TYPE&)*CObList::GetPrev(pos); }
   TYPE& GetAt(POSITION pos)                       {return (TYPE&)*CObList::GetAt(pos);   }
   void SetAt(POSITION pos, TYPE& item)            {CObList::SetAt(pos, &item);           }
   POSITION InsertAfter(POSITION pos, TYPE& item)  {return InsertAfter(pos, &item);       }
   POSITION InsertBefore(POSITION pos, TYPE& item) {return InsertBefore(pos, &item);      }
   POSITION Find(TYPE& item)                       {return CObList::Find(&item);          }
};

///////////////////////////////////////////////////////////////////////////
// TStringArray class
// 
class TStringArray : public CStringArray
{
public:
   bool Contains(const CString& str)
   {
      return Find(str) != -1;
   }
   int Find(const CString& str)
   {
      for (int i = 0; i < GetSize(); i++)
         if (GetAt(i) == str)
            return i;
      return -1;
   }
   int FindNoCase(const CString& str)
   {
      for (int i = 0; i < GetSize(); i++)
         if (GetAt(i).CompareNoCase(str) == 0)
            return i;
      return -1;
   }
   void Sort() // Simple bubble sort - ignores case
   {
      bool bDone = false;
      int j = 0;
      CString Temp;
      while (!bDone) 
      {
         bDone = true;
         j++;
         for (int i = 0; i < GetSize()-j; i++)
            if (GetAt(i).CompareNoCase(GetAt(i+1)) > 0)
            {
               Temp = GetAt(i);
               SetAt(i, GetAt(i+1));
               SetAt(i+1, Temp);
               bDone = false;
            }
      }
   }
   void SortNumbers() // similar to Sort() but keeps numbers in order, 0,1,2,3 .vs. 0,1,10,2,3, etc. 
   {
      bool bDone = false;
      int j = 0;
      CString Temp;
      while (!bDone) 
      {
         bDone = true;
         j++;
         for (int i = 0; i < GetSize()-j; i++)
         {
            CString s0 = GetAt(i);
            CString s1 = GetAt(i+1);
            TCHAR* p;
            double d0 = _tcstod(s0, &p);
            if (*p == _T('\0'))
            {
               double d1 = _tcstod(s1, &p);
               if (*p == _T('\0'))
               {
                  // both are numbers so we can perform a numeric comparison
                  if (d0 > d1)
                  {
                     SetAt(i, s1);
                     SetAt(i+1, s0);
                     bDone = false;
                  }
                  continue;
               }
            }
            // perform an case insensitive alpha comparison
            if (s0.CompareNoCase(s1) > 0)
            {
               SetAt(i, s1);
               SetAt(i+1, s0);
               bDone = false;
            }
         }
      }
   }
};

///////////////////////////////////////////////////////////////////////////
// TArray class
// 
// Improves upon CArray by interally storing pointers to the objects and
// avoiding the moving of objects around in memory, which can result in a 
// memory leak, such as when the object contains a CriticalSection (which 
// cannot be moved in memory without causing a leak in the kernel).
//
template<class TYPE>
class TArray : public CObject
{
public:
// Construction
   TArray();
   ~TArray();

// Attributes
// Attributes
	INT_PTR GetSize() const                              { return m_nSize;                              };
	INT_PTR GetCount() const                             { return m_nSize;                              };
	BOOL IsEmpty() const                                 { return m_nSize == 0;                         };
	INT_PTR GetUpperBound() const                        { return m_nSize-1;                            };
	void SetSize(INT_PTR nNewSize, INT_PTR nGrowBy = -1);

// Operations
	// Clean up
   void FreeExtra();
   void RemoveAll()                                     { SetSize(0, -1);                              };

	// Accessing elements
   const TYPE& GetAt(INT_PTR nIndex) const;
   TYPE& GetAt(INT_PTR nIndex);
   void SetAt(INT_PTR nIndex, const TYPE& newElement);
   const TYPE& ElementAt(INT_PTR nIndex) const;
	TYPE& ElementAt(INT_PTR nIndex);

	// Potentially growing the array
   void SetAtGrow(INT_PTR nIndex, const TYPE& newElement);
   INT_PTR Add(const TYPE& newElement);
	INT_PTR Append(const TArray& src);
	void Copy(const TArray& src);

	// overloaded operator helpers
   const TYPE& operator[](INT_PTR nIndex) const         { return GetAt(nIndex);                        };
	TYPE& operator[](INT_PTR nIndex)                     { return ElementAt(nIndex);                    };

	// Operations that move elements around
   void InsertAt(INT_PTR nIndex, const TYPE& newElement, INT_PTR nCount = 1);
	void RemoveAt(INT_PTR nIndex, INT_PTR nCount = 1);
	void InsertAt(INT_PTR nStartIndex, TArray* pNewArray);

// Implementation
protected:
	TYPE** m_pData;       // the actual array of data
	INT_PTR m_nSize;     // # of elements (upperBound - 1)
	INT_PTR m_nMaxSize;  // max allocated
	INT_PTR m_nGrowBy;   // grow amount

};

template<class TYPE>
AFX_INLINE TYPE& TArray<TYPE>::GetAt(INT_PTR nIndex)
{ 
	ASSERT(nIndex >= 0 && nIndex < m_nSize);
	if(nIndex >= 0 && nIndex < m_nSize)
		return *m_pData[nIndex]; 
	AfxThrowInvalidArgException();		
}

template<class TYPE>
AFX_INLINE const TYPE& TArray<TYPE>::GetAt(INT_PTR nIndex) const
{
	ASSERT(nIndex >= 0 && nIndex < m_nSize);
	if(nIndex >= 0 && nIndex < m_nSize)
		return *m_pData[nIndex]; 
	AfxThrowInvalidArgException();		
}

template<class TYPE>
AFX_INLINE void TArray<TYPE>::SetAt(INT_PTR nIndex, const TYPE& newElement)
{ 
	ASSERT(nIndex >= 0 && nIndex < m_nSize);
	if(nIndex >= 0 && nIndex < m_nSize)
		*m_pData[nIndex] = newElement; 
	else
		AfxThrowInvalidArgException();		
}

template<class TYPE>
AFX_INLINE const TYPE& TArray<TYPE>::ElementAt(INT_PTR nIndex) const
{ 
	ASSERT(nIndex >= 0 && nIndex < m_nSize);
	if(nIndex >= 0 && nIndex < m_nSize)
		return *m_pData[nIndex]; 
	AfxThrowInvalidArgException();		
}

template<class TYPE>
AFX_INLINE TYPE& TArray<TYPE>::ElementAt(INT_PTR nIndex)
{ 
	ASSERT(nIndex >= 0 && nIndex < m_nSize);
	if(nIndex >= 0 && nIndex < m_nSize)
		return *m_pData[nIndex]; 
	AfxThrowInvalidArgException();		
}

template<class TYPE>
void TArray<TYPE>::InsertAt(INT_PTR nIndex, const TYPE& newElement, INT_PTR nCount /*=1*/)
{
	ASSERT_VALID(this);
	ASSERT(nIndex >= 0);    // will expand to meet need
	ASSERT(nCount > 0);     // zero or negative size not allowed

	if(nIndex < 0 || nCount <= 0)
		AfxThrowInvalidArgException();

	if (nIndex >= m_nSize)
	{
		// adding after the end of the array
		SetSize(nIndex + nCount, -1);   // grow so nIndex is valid
	}
	else
	{
		// inserting in the middle of the array
		INT_PTR nOldSize = m_nSize;
		SetSize(m_nSize + nCount, -1);  // grow it to new size
		// destroy intial data before copying over it
		for( int i = 0; i < nCount; i++ )
			delete m_pData[nOldSize + i];
		// shift old data up to fill gap
		::ATL::Checked::memmove_s(m_pData + nIndex + nCount, (nOldSize-nIndex) * sizeof(TYPE*),
			m_pData + nIndex, (nOldSize-nIndex) * sizeof(TYPE*));

		// re-init slots we copied from
		for( int i = 0; i < nCount; i++ )
#pragma push_macro("new")
#undef new
         m_pData[nIndex + i] = new TYPE();
#pragma pop_macro("new")
	}

	// insert new value in the gap
	ASSERT(nIndex + nCount <= m_nSize);
	while (nCount--)
      *m_pData[nIndex++] = newElement;
}

template<class TYPE>
void TArray<TYPE>::RemoveAt(INT_PTR nIndex, INT_PTR nCount)
{
	ASSERT_VALID(this);
	ASSERT(nIndex >= 0);
	ASSERT(nCount >= 0);
	INT_PTR nUpperBound = nIndex + nCount;
	ASSERT(nUpperBound <= m_nSize && nUpperBound >= nIndex && nUpperBound >= nCount);

	if(nIndex < 0 || nCount < 0 || (nUpperBound > m_nSize) || (nUpperBound < nIndex) || (nUpperBound < nCount))
		AfxThrowInvalidArgException();

	// just remove a range
	INT_PTR nMoveCount = m_nSize - (nUpperBound);
	for( int i = 0; i < nCount; i++ )
		delete m_pData[nIndex + i];
	if (nMoveCount)
	{
		::ATL::Checked::memmove_s(m_pData + nIndex, (size_t)nMoveCount * sizeof(TYPE*),
			m_pData + nUpperBound, (size_t)nMoveCount * sizeof(TYPE*));
	}
	m_nSize -= nCount;
}

template<class TYPE>
void TArray<TYPE>::FreeExtra()
{
	ASSERT_VALID(this);

	if (m_nSize != m_nMaxSize)
	{
		// shrink to desired size
#ifdef SIZE_T_MAX
		ASSERT(m_nSize <= SIZE_T_MAX/sizeof(TYPE*)); // no overflow
#endif
		TYPE** pNewData = NULL;
		if (m_nSize != 0)
		{
			pNewData = (TYPE**) new BYTE[m_nSize * sizeof(TYPE*)];
			// copy new data from old
			::ATL::Checked::memcpy_s(pNewData, m_nSize * sizeof(TYPE*),
				m_pData, m_nSize * sizeof(TYPE*));
		}

		// get rid of old stuff (note: no destructors called)
		delete[] (BYTE*)m_pData;
		m_pData = pNewData;
		m_nMaxSize = m_nSize;
	}
}

template<class TYPE>
void TArray<TYPE>::SetSize(INT_PTR nNewSize, INT_PTR nGrowBy = -1)
{
	ASSERT_VALID(this);
	ASSERT(nNewSize >= 0);

	if(nNewSize < 0 )
		AfxThrowInvalidArgException();

	if (nGrowBy >= 0)
		m_nGrowBy = nGrowBy;  // set new size

	if (nNewSize == 0)
	{
		// shrink to nothing
		if (m_pData != NULL)
		{
			for( int i = 0; i < m_nSize; i++ )
            delete m_pData[i];
			delete[] (BYTE*)m_pData;
			m_pData = NULL;
		}
		m_nSize = m_nMaxSize = 0;
	}
	else if (m_pData == NULL)
	{
		// create buffer big enough to hold number of requested elements or
		// m_nGrowBy elements, whichever is larger.
#ifdef SIZE_T_MAX
		ASSERT(nNewSize <= SIZE_T_MAX/sizeof(TYPE *));    // no overflow
#endif
		size_t nAllocSize = __max(nNewSize, m_nGrowBy);
		m_pData = (TYPE**) new BYTE[(size_t)nAllocSize * sizeof(TYPE *)];
		for( int i = 0; i < nNewSize; i++ )
#pragma push_macro("new")
#undef new
         m_pData[i] = new TYPE();
#pragma pop_macro("new")
		m_nSize = nNewSize;
		m_nMaxSize = nAllocSize;
	}
	else if (nNewSize <= m_nMaxSize)
	{
		// it fits
		if (nNewSize > m_nSize)
		{
			// initialize the new elements
			for( int i = 0; i < nNewSize-m_nSize; i++ )
#pragma push_macro("new")
#undef new
            m_pData[m_nSize + i] = new TYPE();
#pragma pop_macro("new")
		}
		else if (m_nSize > nNewSize)
		{
			// destroy the old elements
			for( int i = 0; i < m_nSize-nNewSize; i++ )
				delete m_pData[nNewSize + i];
		}
		m_nSize = nNewSize;
	}
	else
	{
		// otherwise, grow array
		nGrowBy = m_nGrowBy;
		if (nGrowBy == 0)
		{
			// heuristically determine growth when nGrowBy == 0
			//  (this avoids heap fragmentation in many situations)
			nGrowBy = m_nSize / 8;
			nGrowBy = (nGrowBy < 4) ? 4 : ((nGrowBy > 1024) ? 1024 : nGrowBy);
		}
		INT_PTR nNewMax;
		if (nNewSize < m_nMaxSize + nGrowBy)
			nNewMax = m_nMaxSize + nGrowBy;  // granularity
		else
			nNewMax = nNewSize;  // no slush

		ASSERT(nNewMax >= m_nMaxSize);  // no wrap around
		
		if(nNewMax  < m_nMaxSize)
			AfxThrowInvalidArgException();

#ifdef SIZE_T_MAX
		ASSERT(nNewMax <= SIZE_T_MAX/sizeof(TYPE *)); // no overflow
#endif
		TYPE** pNewData = (TYPE**) new BYTE[(size_t)nNewMax * sizeof(TYPE *)];

		// copy new data from old
		::ATL::Checked::memcpy_s(pNewData, (size_t)nNewMax * sizeof(TYPE *),
			m_pData, (size_t)m_nSize * sizeof(TYPE *));

		// construct remaining elements
		ASSERT(nNewSize > m_nSize);
		for( int i = 0; i < nNewSize-m_nSize; i++ )
#pragma push_macro("new")
#undef new
         pNewData[m_nSize + i] = new TYPE();
#pragma pop_macro("new")

		// get rid of old stuff (note: no destructors called)
		delete[] (BYTE*)m_pData;
		m_pData = pNewData;
		m_nSize = nNewSize;
		m_nMaxSize = nNewMax;
	}
}

template<class TYPE>
INT_PTR TArray<TYPE>::Append(const TArray& src)
{
	ASSERT_VALID(this);
	ASSERT(this != &src);   // cannot append to itself
	
	if(this == &src)
		AfxThrowInvalidArgException();

	INT_PTR nOldSize = m_nSize;
	SetSize(m_nSize + src.m_nSize);
	for (INT_PTR i = 0; i<src.m_nSize; i++)
      *m_pData[nOldSize+i] = *src.m_pData[i];
	return nOldSize;
}

template<class TYPE>
void TArray<TYPE>::Copy(const TArray& src)
{
	ASSERT_VALID(this);
	ASSERT(this != &src);   // cannot append to itself

	if(this != &src)
	{
		SetSize(src.m_nSize);
	   for (INT_PTR i = 0; i<src.m_nSize; i++)
         *m_pData[i] = *src.m_pData[i];
	}
}

template<class TYPE>
void TArray<TYPE>::InsertAt(INT_PTR nStartIndex, TArray* pNewArray)
{
	ASSERT_VALID(this);
	ASSERT(pNewArray != NULL);
	ASSERT_VALID(pNewArray);
	ASSERT(nStartIndex >= 0);

	if(pNewArray == NULL || nStartIndex < 0)
		AfxThrowInvalidArgException();

	if (pNewArray->GetSize() > 0)
	{
		InsertAt(nStartIndex, pNewArray->GetAt(0), pNewArray->GetSize());
		for (INT_PTR i = 0; i < pNewArray->GetSize(); i++)
			SetAt(nStartIndex + i, pNewArray->GetAt(i));
	}
}

template<class TYPE>
void TArray<TYPE>::SetAtGrow(INT_PTR nIndex, const TYPE& newElement)
{
	ASSERT_VALID(this);
	ASSERT(nIndex >= 0);
	
	if(nIndex < 0)
		AfxThrowInvalidArgException();

	if (nIndex >= m_nSize)
		SetSize(nIndex+1, -1);

	*m_pData[nIndex] = newElement;
}

template<class TYPE>
INT_PTR TArray<TYPE>::Add(const TYPE& newElement)
{
   INT_PTR nIndex = m_nSize;
	SetAtGrow(nIndex, newElement);
   return nIndex; 
}

template<class TYPE>
TArray<TYPE>::TArray()
{
	m_pData = NULL;
	m_nSize = m_nMaxSize = m_nGrowBy = 0;
}


template<class TYPE>
TArray<TYPE>::~TArray()
{
	ASSERT_VALID(this);

	if (m_pData != NULL)
	{
		for( int i = 0; i < m_nSize; i++ )
         delete m_pData[i];
		delete[] (BYTE*)m_pData;
	}
}

#ifdef _DEBUG
extern "C" void TestTArray();
#endif
