/*
SerialPort.h

Serial Comm functions
*/

#pragma once

#include "Xpert/Win32Fix.h"
#include "GenericPort.h"

#undef DECL
#if defined(NOT_XPERT)
#define DECL
#else
#if defined(_AFXDLL) || defined(XPERT_EXE)
#ifdef UTILS_DLL
#define DECL _declspec(dllexport)
#else
#define DECL _declspec(dllimport)
#endif
#else
#define DECL 
#endif
#endif

/*
SerialPort.h

Serial Comm functions
*/

#pragma once

#include "GenericPort.h"

typedef enum SERIALPOWER {SP_OFF, SP_WAKE, SP_IDLE, SP_LOW, SP_ON};

class DECL CSerialPort : public CGenericPort
{
public:
   CSerialPort();
   ~CSerialPort();

   int GetPortType() { return ptRS232; }; // 0=Unknown, 1=RS232, 2=TCPIP

   CSerialPort(const TCHAR* CommPort);
   CSerialPort& operator=(const CSerialPort& SC);

   // open/close
   bool CloseComm();
   bool OpenComm();
   bool IsOpen() { return (m_hPort != INVALID_HANDLE_VALUE); };

   // configuration
   void SetConfiguration(DWORD BaudRate, BYTE ByteSize, BYTE Parity, BYTE StopBits, BOOL FlowCtrl);
   void SetCommPort(const TCHAR* CommPort) {_tcscpy(m_CommPort, CommPort);};
   void SetBaudRate(const DWORD BaudRate) {m_BaudRate = BaudRate;};
   void SetTimeouts(COMMTIMEOUTS& CommTimeouts) {m_CommTimeouts = CommTimeouts;};
   virtual BOOL SetCommTimeouts(COMMTIMEOUTS& CommTimeouts); 
   SERIALPOWER SetPower(SERIALPOWER State);
   SERIALPOWER GetPower();
   BOOL ResetWake(); // Put the uart back to sleep after it's auto-woken up (restore's the power state)
   // data
   bool GetChar(char * ch) {return GetChar((UCHAR*)ch);}
   bool GetChar(UCHAR * uch);
   bool GetChar(TCHAR * wch);

   // NOTE: Null terminates string, hence Str alloc better be at least 1 > than StrLen!!!
   int GetStr(char * Str, int StrLen) {return GetStr((UCHAR *)Str, StrLen);}
   int GetStr(UCHAR * Str, int StrLen);
   int GetStr(TCHAR * Str, int StrLen);
   
   int GetStr(CString& strResp, DWORD dwTimeout);
   int Get(void* Buffer, int BufferLen);
   int Get(void* Buffer, int BufferLen, DWORD dwTimeout);

   bool Put(const void* Buffer, int BufferLen, DWORD dwTimeout);
   bool PutAndWait(const void* Buffer, int BufferLen, DWORD dwTimeout);
   bool Put(const void* Buffer, int BufferLen);
   bool PutAndWait(const void* Buffer, int BufferLen);
   bool PutChar(const UCHAR uch);
   bool PutChar(const TCHAR wch);

   bool PutStr(const UCHAR * Str, int StrLen);
   bool PutStr(TCHAR * Str, int StrLen);
   bool PutStr(CString Str);

   int  NumberBytesInputBuffer(); // returns number of bytes in input buffer
   bool KeyPressed();             // return true if a key is in rx buffer
   void FlushInput();             // empties rx buffer

   HANDLE GetHandle() {return m_hPort;};

   void WaitOnRx();                // waits (blocks) until a char comes in the serial port - flushes input first
   void WaitForTxEmpty();          // waits (blocks) until transmit buffer empty.
   void WaitForTxEmpty(TCHAR LastCh); // More accurate version of WaitForTxEmpty()

   // When UartMode is set to RS485, the RS485 header on the Xlite is switched into 
   // the circuit, and the behavior of the driver changes, such that RTS drops 
   // automatically after tx buffer is completely emptied on a write. All the app 
   // code must do in this case is set RTS high before write. This behavior supports 
   // quick turn-around in RS485 half-duplex communications. 
   void SetUartMode(int iUartMode);
   int GetUartMode() {return m_UartMode;};

   BOOL SetCommState(LPDCB lpDCB)               { return ::SetCommState(m_hPort, lpDCB);             };
   BOOL GetCommState(LPDCB lpDCB)               { return ::GetCommState(m_hPort, lpDCB);             };
   BOOL GetCommMask(LPDWORD dwEvtMask)          { return ::GetCommMask(m_hPort, dwEvtMask);          };
   BOOL EscapeCommFunction(DWORD dwFunc)        { return ::EscapeCommFunction(m_hPort, dwFunc);      };
   BOOL SetCommBreak()                          { return ::SetCommBreak(m_hPort);                    };
   BOOL ClearCommBreak()                        { return ::ClearCommBreak(m_hPort);                  };
   BOOL GetCommModemStatus(LPDWORD lpModemStat) { return ::GetCommModemStatus(m_hPort, lpModemStat); };
   BOOL SetCommMask(DWORD dwEvtMask)            { return ::SetCommMask(m_hPort, dwEvtMask);          };

// Changes needed to support overlapped I/O in Win32
#ifdef _WIN32_WCE
   BOOL WaitCommEvent(LPDWORD lpEvtMask)        { return ::WaitCommEvent(m_hPort, lpEvtMask, NULL);  };
#else
   BOOL WaitCommEvent(LPDWORD lpEvtMask);
protected:
   BOOL WriteFile(HANDLE hFile, LPCVOID lpBuffer, DWORD nNumberOfBytesToWrite, LPDWORD lpNumberOfBytesWritten, LPOVERLAPPED lpOverlapped);
   BOOL ReadFile(HANDLE hFile, LPVOID lpBuffer,DWORD nNumberOfBytesToRead, LPDWORD lpNumberOfBytesRead, LPOVERLAPPED lpOverlapped);
#endif

protected:
   HANDLE  m_hPort;
   TCHAR   m_CommPort[8];
   DWORD   m_BaudRate;
   BOOL    m_HardwareFlow; // None/Hardware
   BYTE    m_ByteSize;     // number of bits per byte (4-8)
   BYTE    m_Parity;       // 0-4 = no,odd,even,mark,space
   BYTE    m_StopBits;     // 0,1,2 = 1, 1.5, 2

   int     m_UartMode;
   
   SERIALPOWER PowerState;
};